/*-
 * Copyright (c) 2013 Devin Teske <dteske@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/types.h>
#include <stdarg.h>
#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "dprompt.h"
#include "fdpv.h"
#include "string_m.h"

static char dprompt[DPROMPT_MAX] = "";
static char *dprompt_pos = (char *)(0);

/* Display characteristics */
int label_size = 28;	/* Maximum width for labels (file names) */
int pbar_size = 17;	/* How big is the mini-progressbar? */
#define SPIN_SIZE 4
static char spin[SPIN_SIZE + 1] = "/-\\|";
static char *spin_cp = spin;
static int pct_lsize, pct_rsize;
static int done_lsize, done_rsize;
#define PENDING "Pending"
#define PENDING_SIZE 7
static char pending[PENDING_SIZE + 1];
static int pend_lsize, pend_rsize;
unsigned int dheight = 0, dwidth = 0;

/* Function prototypes [for unexported functions] (to prevent warnings) */
char spin_char(void);

/* Returns a pointer to the current spin character in the spin string and
 * advances the global position to the next character for the next call.
 */
char
spin_char(void)
{
	char ch;

	if (!spin_cp) spin_cp = spin;
	ch = *spin_cp;

	/* Advance the spinner to the next char */
	if (++spin_cp >= (spin + SPIN_SIZE)) spin_cp = spin;

	return ch;
}

/* Initialize heights and widths based on various strings and environment
 * variables (such as ENV_USE_COLOR).
 */
void
dprompt_init(void)
{
	char *cp;
	unsigned int len;

	/*
	 * If $USE_COLOR is set and NULL, disable color;
	 * otherwise if set and non-NULL, enable color
	 */
	if ((cp = getenv(ENV_USE_COLOR)) != 0) use_color = ( *cp ? 1 : 0 );

	/* Calculate height of [X]dialog(1) --gauge box */
	dheight = nfiles + 4 + 5;

	/* If using Xdialog(1), adjust accordingly (based on testing) */
	if (use_xdialog) dheight += dheight / 4;

	/* Calculate width of [X]dialog(1) --gauge box */
	dwidth = label_size + pbar_size + 9;

	/* If prefix or append strings are longer, adjust accordingly */
	len = dprompt_longestline(pprompt);
	if ((len + 4) > dwidth) dwidth = len + 4;
	len = dprompt_longestline(aprompt);
	if ((len + 4) > dwidth) dwidth = len + 4;

	/* Calculate left/right portions of % and Done */
	pct_lsize = (pbar_size - 4) / 2; /* -4 == printf("%-3s%%", pct) */
	pct_rsize = pct_lsize;
	done_lsize = (pbar_size - 4) / 2; /* -4 == "Done" */
	done_rsize = done_lsize;
	/* If not evenly divisible by 2, increment the right-side */
	if ((pct_rsize + pct_rsize + 4) != pbar_size) pct_rsize++;
	if ((done_rsize + done_lsize + 4) != pbar_size) done_rsize++;

	/* Initialize "Pending" text */
	if (pbar_size < PENDING_SIZE)
	{
		pend_lsize = pbar_size;
		pend_rsize = 0;
		*(pending + pbar_size) = '\0';
	}
	else
	{
		*pending = '\0';
		strcat(pending, PENDING);
		/* Calculate left/right portions of mini-progressbar */
		pend_lsize = (pbar_size - PENDING_SIZE) / 2;
		pend_rsize = pend_lsize;
		/* If not evenly divisible by 2, increment the right-side */
		if ((pend_rsize + pend_lsize + 7) != pbar_size) pend_rsize++;
	}
}

/* Clear the [X]dialog(1) `--gauge' prompt buffer. */
void
dprompt_clear(void)
{
	*dprompt = '\0';
	dprompt_pos = dprompt;
}

/* Append to the [X]dialog(1) `--gauge' prompt buffer. Syntax is like printf(3)
 * and returns the number of bytes appended to the buffer.
 */
int
dprompt_add(const char *format, ...)
{
	va_list ap;
	int len;

	va_start(ap, format);
	len = vsnprintf(dprompt_pos,
		(size_t)(DPROMPT_MAX - (dprompt_pos - dprompt)),
		format, ap);
	va_end(ap);
	if (len == -1)
	{
		fprintf(stderr, "%s:%d: Oops, dprompt buffer overflow\n",
		        __FILE__, __LINE__);
		exit(EXIT_FAILURE);
	}

	dprompt_pos += len;
	return len;
}

/* Process the file_node linked-list of named file-descriptors, re-generating
 * the [X]dialog(1) `--gauge' prompt text for the current state of transfers.
 */
void
dprompt_recreate(file_node *file_list, file_node *curfile, int pct)
{
	file_node *fp;
	u_int8_t after_curfile = ( curfile ? 0 : 1 );
	char pbar[pbar_size + 16]; /* +15 for optional color */

	/*
	 * Create the mini-progressbar
	 */
	snprintf(pbar, pbar_size + 1, "%*s%3u%%%*s",
	         pct_lsize, "", pct, pct_rsize, "");
	if (use_color)
	{
		int pwidth;
		char pbar_fill[sizeof(pbar)];
		char pbar_cap[sizeof(pbar)];

		/* Calculate the fill-width of progressbar */
		pwidth = pct * pbar_size / 100;
		/* Round up based on one-tenth of a percent */
		if ((pct * pbar_size % 100) > 50) pwidth++;

		/*
		 * Make two copies of pbar. Make one represent the fill and the
		 * other the remainder (cap). We'll insert the ANSI delimiter
		 * in between.
		 */
		*pbar_fill = '\0';
		*pbar_cap = '\0';
		strncat(pbar_fill, (const char *)(pbar), dwidth);
		*(pbar_fill + pwidth) = '\0';
		strncat(pbar_cap, (const char *)(pbar+pwidth), dwidth);

		/* Finalize the mini [color] progressbar */
		snprintf(pbar, sizeof(pbar), "\\Zb\\Zr\\Z4%s%s%s\\Zn",
		         pbar_fill, "\\ZR", pbar_cap);
	}

	/*
	 * Build the prompt text
	 */
	dprompt_clear();
	for (fp = file_list; fp; fp = fp->next)
	{
		char *name = fp->name;
		char flabel[label_size + 1];

		/* Are we before the current file still? */
		if (fp == curfile) after_curfile = 1;

		/* If name is too long, add an ellipsis */
		if (snprintf(flabel, label_size + 1, "%s", name) > label_size)
			sprintf(flabel + label_size - 3, "...");

		/*
		 * Append the label (processing the current file differently)
		 */
		if (fp == curfile && pct < 100)
		{
			char *cp;

			/*
			 * Add an ellipsis to current file name if it will fit.
			 * There may be an ellipsis already from truncating the
			 * label (in which case, we already have one).
			 */
			cp = flabel + strlen(flabel);
			if (cp < (flabel + label_size))
				snprintf(cp, label_size - (cp - flabel) + 1,
				         "...");

			/* Append label (with spinner and optional color) */
			dprompt_add("%s%-*s%s %c",
			            (use_color ? "\\Zb" : ""),
			            label_size, flabel,
			            (use_color ? "\\ZB" : ""),
			            spin_char());
		}
		else
			dprompt_add("%-*s %s", label_size, flabel, " ");

		/*
		 * Append pbar/status (processing the current file differently)
		 */
		if (fp == curfile && pct < 100)
			dprompt_add(" [%s]\\n", pbar);
		else if (fp != curfile && after_curfile)
			dprompt_add(" [%-*s%s%-*s]\\n",
			            pend_lsize, "", pending, pend_rsize, "");
		else
			dprompt_add(" [%s%-*s%s%-*s%s]\\n",
			            (use_color ?  "\\Zb\\Zr\\Z2" : ""),
			            done_lsize, "", "Done", done_rsize, "",
			            (use_color ?  "\\Zn" : ""));
	}

	/* Xdialog(1) requires newlines (a) escaped and (b) in triplicate */
	if (use_xdialog)
	{
		size_t len;

		/* Replace `\n' with `\n\\n\n' in dprompt */
		len = strlen(dprompt);
		len += strcount(dprompt, "\\n") * 2;
		if (len > DPROMPT_MAX)
		{
			fprintf(stderr, "%s:%d: Oops, dprompt buffer overflow"
			        " (%i > %i)\n", __FILE__, __LINE__,
			        len, DPROMPT_MAX);
			exit(EXIT_FAILURE);
		}
		if (replaceall(dprompt, "\\n", "\n\\n\n") < 0)
		{
			fprintf(stderr, "%s:%d: replaceall: unknown error\n",
			        __FILE__, __LINE__);
			exit(EXIT_FAILURE);
		}
	}
}

/* Print the [X]dialog(1) `--gauge' prompt text to a buffer. */
int
dprompt_sprint(char * restrict str, const char *prefix, const char *append)
{
	return snprintf(str, DPROMPT_MAX, "%s%s%s",
	                prefix ? prefix : "", dprompt, append ? append : "");
}

/* Print the [X]dialog(1) `--gauge' prompt text to file descriptor fd (could
 * be STDOUT_FILENO or a pipe(2) file descriptor to actual [X]dialog(1)).
 */
void
dprompt_dprint(int fd, const char *prefix, const char *append, int overall)
{
	dprintf(fd, "XXX\n%s%s%s\nXXX\n%i\n",
	        prefix ? prefix : "", dprompt, append ? append : "", overall);
	fsync(fd);
}

/* Returns the length in bytes of the longest line in buffer pointed to by
 * prompt pointer. Takes newlines and escaped newlines into account. Also
 * discounts dialog(1) color escape codes if enabled (via `use_color' global).
 */
unsigned int
dprompt_longestline(const char *prompt)
{
	const char *p = prompt;
	int n = 0, longest = 0;
	u_int8_t backslash = 0;

	/* Parameters is required */
	if (!prompt) return 0;

	/* Loop until the end of the string */
	while (*p) {
		/* check for backslash character */
		if (*p == '\\')
		{
			/* if second backslash, count as a single-char */
			if (!(backslash = ++backslash % 2)) n++;
		}
		else if (backslash)
		{
			if (*p == 'n') /* new line */
			{
				if (n > longest) longest = n;
			}
			else if (use_color && *p == 'Z')
			{
				if (*++p) p++;
				backslash = 0;
				continue;
			}
			backslash = n = 0;
		}
		else
			n++;
		p++;
	}
	if (n > longest) longest = n;

	return longest;
}
